;Treibt 8x2 Anzeige in dem Labornetzteil 0-16V 0-5A 
;Gemessen werden die Ausgangsspannung und der Ausgangsstrom
;Anzeige in der Form:
;"U:AA,AAV"
;"I:B,BBBA"
;;--------------------------------------------------------------------------
;; Connections
;;  Messeingang Spannung - AN1, Pin3
;;  Messeingang Strom - AN0, Pin2
;;  Eingang Rohwertanzeige - RA4, Pin6 (0: Rohe ADC-Werte, 1: umgerechnete V/A-Werte)
;;  LCD Daten - RC0-RC7  = Pin11-18 
;;  LCD Control RS - RB2, Pin23
;;  LCD Control R/W - RB1, Pin22
;;  LCD Control E - RB0, Pin21 
;;--------------------------------------------------------------------------
;;
;; Referenzspannung fr den ADC ist 4,096V an VREF+, Pin5
;; Clock wird durch externen RC-Oszillator gemacht, keine genaue time-Basis notwendig. Takt ist ca. 4MHz.
;;
;;Config-register sollte bei Programmierung auf 11 1111 0011 0011 (3F33) stehen
;;(cp=11: code protection off, debug=1: disabled, wrt=1: writable, cpd=1: cp off, lvp=0: GIO enabled, 
;;boren=0: bor disabled, pwrten=0: power up timer on, wdten=0: kein watchdog timer, fosc=11: externer RC oszillator)


	processor  16F870
	radix  DEC
	__CONFIG 0x3F33

;Register Adressen
INDF        EQU   0x00
TMR0        EQU   0x01
STATUS      EQU   0x03
FSR         EQU   0x04
PORTA       EQU   0x05
PORTB       EQU   0x06
PORTC       EQU   0x07
INTCON      EQU   0x0B
PIR1	    EQU   0x0C
TMR1L       EQU   0x0E
TMR1H       EQU   0x0F
T1CON       EQU   0x10
TMR2        EQU   0x11
T2CON       EQU   0x12
CP1CON		EQU	  0x17
RCSTA		EQU   0x18
ADRESH      EQU   0x1E
ADCON0      EQU   0x1F
OPTIONR     EQU   0x81
TRISA       EQU   0x85
TRISB       EQU   0x86
TRISC       EQU   0x87
PIE1        EQU   0x8C
PIE2	    EQU   0x8D
PR2         EQU   0x92
ADRESL      EQU   0x9E
ADCON1      EQU   0x9F
EEDATA	    EQU   0x10C
EEADR	    EQU   0x10D
EECON1	    EQU	  0x18C
EECON2      EQU   0x18D

_C      	EQU 	0		;Carry-Bit
_Z      	EQU   	2		;Zero-Bit
_W			EQU		0		;Destination W bei z.B. INCF
_R			EQU		1		;Destination selbes Register bei z.B. INCF 

#define	_BANK0	BCF STATUS,5	;Umschalten auf Bank0
#define	_BANK1	BSF STATUS,5	;Umschalten auf Bank1
#define	_iRAW	PORTA,4			;Eingang Rohwertanzeige (0: Rohwerte, 1: V/A-Werte)
#define	_dRS	PORTB,2			;RS an Display
#define	_dRW	PORTB,1			;RW an Display
#define	_dE		PORTB,0			;E an Display
#define	_dDAT	PORTC			;Data 0..7 an Display
#define _ADCGO	ADCON0,2		;Go-Bit fr ADC

;Kalibrierungsdaten fuer die Messung, bzw. Berechnung der Geraden
;Werte sind durch Kalibrierungsmessungen ermittelt und mit Kalibrierung.ods berechnet - siehe dort 
VStg		EQU		57		;Skalierte Steigung fuer Spannungswertberechnung 
VOff		EQU		7		;Offset fuer Spannung - eigentlich 6, aber zustzlich +1 wg. Rundung
AStg		EQU		46		;Skalierte Steigung fuer Stromwertberechnung
AOff		EQU		28		;Offset fuer Strom - eigentlich -29, aber zustzlich +1 wg. Rundung (Offset ist negativ, d.h. in Stromberechnung Subtraktion statt Addition)


;Memory
MEM         EQU   0x20
Disp_D      EQU   MEM+0		;Datenwort fuer Displayanzeige
Disp_C	    EQU   MEM+1		;Verschiedene Kontrollbits fuer Displayanzeige
Char1       EQU   MEM+2		;Displaybuffer
Char2       EQU   MEM+3
Char3       EQU   MEM+4
Char4       EQU   MEM+5
Char5       EQU   MEM+6
Char6       EQU   MEM+7
Char7       EQU   MEM+8
Char8       EQU   MEM+9
Char9       EQU   MEM+10
Char10      EQU   MEM+11
Char11      EQU   MEM+12
Char12      EQU   MEM+13
Char13      EQU   MEM+14
Char14      EQU   MEM+15
Char15      EQU   MEM+16
Char16      EQU   MEM+17
n           EQU   MEM+18	;Allgemeiner Zaehler
DelayCnt    EQU   MEM+19	;Parameter, der angibt, wielange Delayroutine benoetigen soll
UH			EQU	  MEM+20	;ADC-Messwert der Spannung, High und Low-Byte
UL			EQU	  MEM+21
IH			EQU	  MEM+22	;ADC-Messwert des Stroms, High und Low-Byte
IL			EQU	  MEM+23
Erg1		EQU	  MEM+24	;16-Bit Ergebnis
Erg0		EQU	  MEM+25
OpA1		EQU	  MEM+26	;16-Bit Operand
OpA0		EQU	  MEM+27
OpB1		EQU	  MEM+28	;16-Bit Operand
OpB0		EQU	  MEM+29
Mx			EQU	  MEM+30	;Zhler fr Multiplikation
My			EQU	  MEM+31	;Zhler fr Division
M1			EQU	  MEM+32	;16-Bit Hilfs-Variable fr Divison
M0			EQU	  MEM+33
x           EQU   MEM+34	;Allgemeiner Zaehler  fr Delay-Routine


;******* Programm Start **************************************************
	ORG	0
 		GOTO	main

;******* Interrupt Service Routine ***************************************
	ORG	4


#include mathe.asm			;Generelle mathematische Funktionen einbinden	


;***********Main ********************************************
main	;Prozessor Init
		CLRF  	PORTA
		CLRF  	PORTB
		CLRF  	PORTC
		CLRF 	INTCON		;Alle Interrupts disabled	
		CLRF 	T1CON		;TMR1 ausschalten
		CLRF 	T2CON		;TMR2 ausschalten
		CLRF	CP1CON		;CCP ausschalten
		CLRF	RCSTA		;UART ausschalten
		MOVLW 	0x41		;ADC initialisieren: Fosc/8, Eingeschaltet aber noch nicht aktiv
		MOVWF 	ADCON0
		_BANK1				;Umschalten auf Bank 1
		MOVLW 	0x87		;Options: Prescale 1:256 auf TMR0			
		MOVWF 	OPTIONR	
		MOVLW 	0xFF		
		MOVWF 	TRISA		;PortA: Alles Inputs
		MOVLW 	0xF8
		MOVWF  	TRISB		;PortB - 0,1,2 sind Output, Rest Input
		CLRF 	TRISC		;PortC - Alles Output
		CLRF 	PIE1		;PIE Inits: Alle Interrupts off
		CLRF  	PIE2
		MOVLW 	0x85
		MOVWF 	ADCON1		;ADCON1: Right Justified, AN0 und AN1 enabled, VRef kommt von VRef+
		_BANK0 				;Bank 0

		;Display Init
		MOVLW 	.255		;Warte 2,05*10 = 20,5ms
		MOVWF 	DelayCnt
		CALL 	delayx
		CALL 	delayx
		CALL 	delayx
		CALL 	delayx
		CALL 	delayx
		CALL 	delayx
		CALL 	delayx
		CALL 	delayx
		CALL 	delayx
		CALL 	delayx
		BCF   	Disp_C,0	;Kontrolldaten werden gesendet
		MOVLW 	0x38		;Sende erstes Function Set Wort		
		MOVWF 	Disp_D
		CALL 	DisplayWriteWord
		CALL 	delayx		;Warte 6ms
		CALL 	delayx
		CALL 	delayx
		CALL 	DisplayWriteWord	;Sende zweites Function Set Wort
		CALL 	delayx		;Warte 2ms
		CALL 	DisplayWriteWord	;Sende drittes Function Set Wort
		CALL 	delayx		;Warte 2ms
		CALL 	DisplayWriteWord	;Sende gueltiges Function Set Wort (8Bit IF, 2 line display, 5x7 dots) 
		CALL 	delayx		;Warte 2ms
		MOVLW 	0x01		;Display Clear	
		MOVWF 	Disp_D
		CALL 	DisplayWriteWord	
		CALL 	delayx		;2ms Warten
		MOVLW 	0x06		;Cursor increase, display not shifted	
		MOVWF 	Disp_D
		CALL 	DisplayWriteWord	
		CALL 	delayx		;2ms Warten
		MOVLW 	0x0C		;Display on, cursor off, blinking off	
		MOVWF 	Disp_D
		CALL 	DisplayWriteWord	
		CALL 	delayx		;2ms Warten	
		;Display Init beendet

		;Fixe Stellen in der Anzeige initialisieren
		;"U:AA,AAV" bzw. "U:AA AA "
		;"I:B,BBBA" bzw. "I:B BBB "
		MOVLW 	0x3A		;An Stelle 2 und 10 wird immer ein ":" angezeigt
		MOVWF 	Char2
		MOVWF 	Char10
		MOVLW 	0x55		;An Stelle 1 wird immer ein "U" angezeigt
		MOVWF 	Char1
		MOVLW 	0x49		;An Stelle 9 wird immer ein "I" angezeigt
		MOVWF 	Char9

		CLRF	TMR0

Loop	MOVF	TMR0,_W		;TMR0 prfen
		BTFSS	STATUS,_Z	;Ist TMR0=0?
		GOTO	$-2			;Nein: Weiterwarten / Ja: Messungen, Umrechnung und Anzeige starten

		;********* Spannungsmessung ******************
		MOVLW 	0x49		;ADC auf Kanal 1 schalten
		MOVWF 	ADCON0
		MOVLW 	.123		;Warte 1ms acquisition time
		MOVWF 	DelayCnt
		CALL 	delayx
		BSF   	_ADCGO		;Starte ADC Conversion
		BTFSC 	_ADCGO		;Pruefe, ob Konvertierung beendet (GO=0)
		GOTO  	$-1			;Wenn nein, weiterwarten
		MOVF  	ADRESH,_W	;Wenn ja, ADRESH...	
		MOVWF 	UH			;...nach UH speichern
		_BANK1			 	;Umschalten auf Bank 1
		MOVF  	ADRESL,_W	;ADRESL...
		_BANK0			 	;Umschalten auf Bank 0
		MOVWF 	UL			;...nach UL speichern
	

		;********* Strommessung ******************
		MOVLW 	0x41		;ADC auf Kanal 0 schalten
		MOVWF 	ADCON0
		MOVLW 	.123		;Warte 1ms acquisition time
		MOVWF 	DelayCnt
		CALL 	delayx
		BSF   	_ADCGO		;Starte ADC Conversion
		BTFSC 	_ADCGO		;Pruefe, ob Konvertierung beendet (GO=0)
		GOTO  	$-1			;Wenn nein, weiterwarten
		MOVF  	ADRESH,_W	;Wenn ja, ADRESH...	
		MOVWF 	IH			;...nach IH speichern
		_BANK1			 	;Umschalten auf Bank 1
		MOVF  	ADRESL,_W	;ADRESL...
		_BANK0			 	;Umschalten auf Bank 0	
		MOVWF 	IL			;...nach IL speichern


		;Standardcharacter fr Rohwertanzeige
		MOVLW 	0x20		;An Stelle 5 und 8 wird im Rohwertmodus ein " " angezeigt, das wird spter wieder gendert, wenn kein Rohwert-Modus sein soll
		MOVWF 	Char5
		MOVWF 	Char8
		;Spannungsberechnung und Zifferngenerierung
		MOVF	UH,_W		;OpA=U
		MOVWF	OpA1
		MOVF	UL,_W
		MOVWF	OpA0
		BTFSS	_iRAW		;Ist Raw-Pin=0 (Rohwertanzeige)?
		GOTO	UEx			;Ja: Umrechnung berspringen - Rohwert ausgeben
		MOVLW	VStg		;Nein: OpB=VStg
		MOVWF	OpB0
		CALL	MvEAB221	;Erg=U*VStg (Solange VStg<=64 ist, gibt es kein Risiko eines Ueberlaufs)
		CALL	SRvE2		;Erg=Erg/16 - De-Skalierung um 32/2 = 16 = 4 mal /2 - nur die Hlfte de-skalieren wg. Rundung
		CALL	SRvE2
		CALL	SRvE2
		CALL	SRvE2
		CLRF	OpA1		;OpA=VOff
		MOVLW	VOff
		MOVWF	OpA0
		CALL	AvEEA222	;Erg=Erg+VOff+1 (+1 fr Rundung: +1 und dann /2 entspricht Rundung)
		CALL	SRvE2		;Erg=Erg/2 - Restliche De-Skalierung um 2 auf insgesdamt 32
		MOVF	Erg1,_W		;OpA=Erg
		MOVWF	OpA1
		MOVF	Erg0,_W
		MOVWF	OpA0
		;Character fr Spannungsanzeige (berschreibt vorher eingestellte Rohanzeige)
		MOVLW 	0x2C		;An Stelle 5 wird ein "," angezeigt
		MOVWF 	Char5
		MOVLW 	0x56		;An Stelle 8 wird immer ein "V" angezeigt
		MOVWF 	Char8		
UEx		MOVLW	0x30		;Leading-Zero-Elimination vorbereiten
		MOVWF	x
		CALL	C4vA2		;Extrahiert 1000er Character aus OpA und setzt Character
		SUBWF	x,_R		;x=0x30-W
		BTFSC	STATUS,_Z	;=0 (W=0x30 (="0"))?
		MOVLW	0x20		;Ja: "0" durch " " ersetzen	
		MOVWF 	Char3
		CALL	C3vA2		;Extrahiert 100er Character aus OpA	und setzt Character	
		MOVWF 	Char4
		CALL	C2vA1		;Extrahiert 10er Character aus OpA und setzt Character		
		MOVWF 	Char6
		CALL	C1vA1		;Extrahiert 1er Character aus OpA und setzt Character		
		MOVWF 	Char7		

		;Standardcharacter fr Rohwertanzeige
		MOVLW 	0x20		;An Stelle 12 und 16 wird im Rohwertmodus ein " " angezeigt, das wird spter wieder gendert, wenn kein Rohwert-Modus sein soll
		MOVWF 	Char12
		MOVWF 	Char16
		;Stromberechnung und Zifferngenerierung
		MOVF	IH,_W		;OpA=I
		MOVWF	OpA1
		MOVF	IL,_W
		MOVWF	OpA0
		BTFSS	_iRAW		;Ist Raw-Pin=0 (Rohwertanzeige)?
		GOTO	IEx			;Ja: Umrechnung berspringen - Rohwert ausgeben
		MOVLW	AStg		;OpB=AStg
		MOVWF	OpB0
		CALL	MvEAB221	;Erg=I*AStg (Solange OStg<=64 ist, gibt es kein Risiko eines Ueberlaufs)
		CALL	SRvE2		;Erg=Erg/8 - De-Skalierung um 8/2 = 4 = 2 mal /2 - nur die Hlfte de-skalieren wg. Rundung
		CALL	SRvE2
		CLRF	OpA1		;OpA=AOff
		MOVLW	AOff
		MOVWF	OpA0
;		CALL	AvEEA222	;Offset ist positov: Erg=Erg+AOff+1 (+1 fr Rundung: +1 und dann /2 entspricht Rundung)
		CALL	SvEEA221	;Offset ist negativ: Erg=Erg-AOff+1 (+1 fr Rundung: +1 und dann /2 entspricht Rundung)
		INCF	Erg1,_W		;Bei negativen Offset muss auf Borrow geprft werden, d.h. wenn Erg1=255 ist, fand in Subtraction ein borrow statt
		BTFSS	STATUS,_Z	;Ist Zero=0, d.h. Erg1 war vor INCF 255
		GOTO	$+3			;Nein: In Subtraction hat kein Borrow stattgefunden - direkt weiter mit Auswertung
		CLRF	Erg1		;Ja: In Subtraction hat Borrow stattgefunden, d.h. Ergebnis ist "negativ" - Ergebnis auf 0
		CLRF	Erg0
		CALL	SRvE2		;Erg=Erg/2 - Restliche De-Skalierung um 2 auf insgesdamt 8
		MOVF	Erg1,_W		;OpA=Erg
		MOVWF	OpA1
		MOVF	Erg0,_W
		MOVWF	OpA0
		;Character fr Sromanzeige (berschreibt vorher eingestellte Rohanzeige)
		MOVLW 	0x2C		;An Stelle 12 wird ein "," angezeigt
		MOVWF 	Char12
		MOVLW 	0x41		;An Stelle 16 wird immer ein "A" angezeigt
		MOVWF 	Char16
IEx		CALL	C4vA2		;Extrahiert 1000er Character aus OpA und setzt Character		
		MOVWF 	Char11
		CALL	C3vA2		;Extrahiert 100er Character aus OpA	und setzt Character	
		MOVWF 	Char13
		CALL	C2vA1		;Extrahiert 10er Character aus OpA und setzt Character		
		MOVWF 	Char14
		CALL	C1vA1		;Extrahiert 1er Character aus OpA und setzt Character		
		MOVWF 	Char15	

		CALL 	DisplayWriteLine	;Display neu beschreiben
		GOTO	Loop		;Auf nchsten Zhlerberlauf warten



;******* DelayX *******************************************************
delayx						;kurzes delay mittels for-Schleife und NOPS
		MOVLW 	.1			;Delay = 4 + DelayCnt*8 + 7
		MOVWF 	x			;Prolog: CALL(2) + MOV * MOV = 4 Zyklen
dloop	MOVF 	DelayCnt,_W	;Schleife: MOV + SUB + BTFSC(2) + NOP + INC + GOTO(2) = 8 Zyklen
		SUBWF 	x,_W		;Epilog: MOV + SUB + BTFSC(1) + GOTO(2) + RET(2) = 7 Zyklen
		BTFSC 	STATUS,_C
		GOTO  	edelay
		NOP
		INCF  	x,_R
		GOTO  	dloop
edelay  RETLW 	.1


;******* DisplayWriteWord *********************************************
DisplayWriteWord			;Zu Schreibende data liegt in Disp_D, Bit 0 von Disp_C entscheidet wie geschrieben wird
		BTFSS 	Disp_C,0	;check, ob Contol-Wort (0) oder Datenwort (1) zu schreiben
		GOTO 	$+3			;wenn Controlwort: RS=0
		BSF   	_dRS		;wenn Datenwort: RS=1
        GOTO 	$+2
		BCF   	_dRS
	    BCF   	_dRW		;R/W=0 -> Write Modus
		BSF   	_dE			;E=1 -> Prepare Write
		MOVF 	Disp_D,_W	;Datum auf Datenbus ausgeben
		MOVWF 	_dDAT
		BCF   	_dE			;E=0 -> Write
		RETLW 	.1

;******* DisplayWriteLine ********************************************
DisplayWriteLine
		BCF   	Disp_C,0	;Zuerst werden Kontrolldaten gesendet
		MOVLW 	0x01		;Clear Display
		MOVWF 	Disp_D
		CALL 	DisplayWriteWord
		MOVLW 	.255		;Warte 2,05 ms
		MOVWF 	DelayCnt
		CALL 	delayx
		CLRF  	n			;Schleife vorbereiten
		MOVLW 	.11			;Warteparameter fuer Delay in Schleife vorbereiten
		MOVWF 	DelayCnt
wchr	BSF   	Disp_C,0	;Ab jetzt werden datenworte geschrieben
		MOVLW 	Char1		;Indirekte Adressierung Platziere Pointer auf Char1
		ADDWF 	n,_W		;Erhoehe pointer auf Charx
		MOVWF 	FSR		
		MOVF  	INDF,_W		;Charx ist jetzt indirect ansprechbar ueber INDF
		MOVWF 	Disp_D		;Charx nach Disp_D schieben
		CALL  	DisplayWriteWord
		CALL 	delayx		;Warte 99 us
		INCF  	n,1
		MOVLW 	0x08
		SUBWF 	n,_W		;Pruefe, ob n den Wert 8 erreicht hat
		BTFSS 	STATUS,_Z
		GOTO  	chke		;Wenn nein, naechster Char uebertragen
		BCF   	Disp_C,0	;Wenn ja, DDRAM Adresse auf 0x40 (hintere 8 Zeichen) setzen
		MOVLW 	0xC0		;Adresse auf 0x40 setzen
		MOVWF 	Disp_D
		CALL 	DisplayWriteWord
		CALL 	delayx		;Warte 99 us	
chke	MOVLW 	0x10		
		SUBWF 	n,_W		;Pruefe, ob n den Wert 16 erreicht hat
		BTFSS 	STATUS,_C
		GOTO  	wchr		;Wenn nein, naechtster Char uebertragen
		RETLW 	.1			;Wenn ja, Schluss


		END

