;;******* MvEAB221 *******************************************************
;Funktion: Multipliziert 16bit mit 8bit zu 16bit Ergebnis. 
;Funktion: Erg[2]=OpA[2]*OpB[1]
;Bentigte Subroutinen: AvEEA222, SLvA2
;Bentigte Variablen: Erg[2], OpA[2], OpB[1], Mx
;Bentigte Register: STATUS
;Unverndert bleibt: nichts
;Verndert werden  : Erg[2], OpA[2], OpB[1], Mx, STATUS
MvEAB221	CLRF	Erg0		;Clear Lowbyte des Ergebnisses
		CLRF	Erg1		;Clear Highbyte des Ergebnisses
		MOVLW   .8			;8-fach Schleife
		MOVWF   Mx
		RRF		OpB0,1		;Right shift OpB nach carry
		BTFSC	STATUS, 0	;Check geshiftetes Bit
		CALL	AvEEA222		;Wenn 1, Erg=Erg+OpA
		CALL	SLvA2		;Left shift OpA	
		DECFSZ	Mx,1		;Schleife runterzhlen
		GOTO	$-5			;Neue Schleife
		RETURN


;;******* AvEEA222 *******************************************************
;Funktion: Addiert 16bit mit 16bit zu 16bit Ergebnis. 
;Funktion: Erg[2]=Erg[2]+OpA[2]
;Bentigte Subroutinen: keine
;Bentigte Variablen: OpA[2], Erg[2]
;Bentigte Register: STATUS
;Unverndert bleibt: OpA[2]
;Verndert werden  : Erg[2], STATUS 
AvEEA222	MOVF	OpA1,0
		ADDWF	Erg1,0		
		MOVWF	Erg1		;Erg1=Erg1+OpA1
		MOVF	OpA0,0
		ADDWF	Erg0,0
		MOVWF	Erg0		;Erg0=Erg0+OpA0
		BTFSC	STATUS,0	;Pruefe, ob Ueberlauf
		INCF	Erg1,1		;berlauf: Erhoehe Erg1
		RETURN			


;;******* SLvA2 *******************************************************
;Funktion: Shifted einen 16bit Wert um eins nach links. MSB geht nach Carry, LSB wird mit 0 aufgefuellt
;Funktion: OpA[2]=OpA[2]*2
;Bentigte Subroutinen: keine
;Bentigte Variablen: OpA[2]
;Bentigte Register: STATUS
;Unverndert bleibt: nichts
;Verndert werden  : OpA[2], STATUS 
SLvA2	BCF		STATUS,_C	;mit 0 auffllen	
		RLF		OpA0,1		;Left shift lowbyte, oberstes Bit geht nach Carry
		RLF		OpA1,1		;Left shift highbyte, unterstes Bit kommt von Carry
		RETURN


;;******* SRvE2 *******************************************************
;Funktion: Shifted einen 16bit Wert um eins nach rechts. MSB geht nach Carry, LSB wird mit 0 aufgefuellt
;Funktion: Erg[2]=Erg[2]/2
;Bentigte Subroutinen: keine
;Bentigte Variablen: Erg[2]
;Bentigte Register: STATUS
;Unverndert bleibt: nichts
;Verndert werden  : Erg[2], STATUS 
SRvE2	BCF		STATUS,0	;mit 0 auffllen	
		RRF		Erg1,1		;Right shift highbyte, unterstes Bit geht nach Carry
		RRF		Erg0,1		;Right shift lowbyte, oberstes Bit kommt von Carry
		RETURN		


;;****** C4vA2 *******************************************************
;Funktion: Extrahiert den 1000er Character aus OpA[2] und gibt diesen als Chr-Code fr Displays aus
;Funktion: W=CHRCODE(OpA[2]/1000), OpA[2]=Rest
;Bentigte Subroutinen: DvEAB222, VvMB22, SvMMB222
;Bentigte Variablen: Erg[2], OpA[2], OpB[2], M[2], My
;Bentigte Register: STATUS
;Unverndert bleibt: keine
;Verndert werden  : Erg[2], OpA[2], OpB[2], M[2], My, STATUS
C4vA2	MOVLW	0xE8		;1000 nach OpB
		MOVWF	OpB0
		MOVLW	0x03
		MOVWF	OpB1
		CALL 	DvEAB222	;Erg=OpA/1000; OpA=Rest
		MOVF	Erg0,0		;Ergebnis nach W
		ADDLW	0x30		;Char generieren (im Zeichensatz ist "0" an Position 0x30 bis "9" an Position 0x39)
		RETURN		
	

;;****** C3vA2 *******************************************************
;Funktion: Extrahiert den 100er Character aus OpA[2] und gibt diesen als Chr-Code fr Displays aus
;Funktion: W=CHRCODE(OpA[2]/100), OpA[2]=Rest : User muss darauf achten, dass OpA[2] nicht grer als 999 ist
;Bentigte Subroutinen: DvEAB222, VvMB22, SvMMB222
;Bentigte Variablen: Erg[2], OpA[2], OpB[2], M[2], My
;Bentigte Register: STATUS
;Unverndert bleibt: keine
;Verndert werden  : Erg[2], OpA[2], OpB[2], M[2], My, STATUS
C3vA2	MOVLW	.100		;100 nach OpB
		MOVWF	OpB0
		CLRF	OpB1
		CALL 	DvEAB222	;Erg=OpA/100; OpA=Rest
		MOVF	Erg0,0		;Ergebnis nach W
		ADDLW	0x30		;Char generieren (im Zeichensatz ist "0" an Position 0x30 bis "9" an Position 0x39)
		RETURN		
		

;;***** C2vA1 *******************************************************
;Funktion: Extrahiert den 10er Character aus OpA0 und gibt diesen als Chr-Code fr Displays aus
;Funktion: W=CHRCODE(OpA[1]/10), OpA0=Rest : User muss darauf achten, dass OpA0 nicht grer als 99 ist
;Bentigte Subroutinen: keine
;Bentigte Variablen: Mx, OpA[1] 
;Bentigte Register: STATUS
;Unverndert bleibt: keine
;Verndert werden  : Mx, OpA[1], STATUS
C2vA1		CLRF  	Mx			
		MOVLW 	.10
		SUBWF 	OpA0,0		;W=OpA0-10
		BTFSS 	STATUS, 0	;Wurde 0 unterschritten (war letztes OpA0<10)
		GOTO 	$+4			;Wenn ja, unten weiter
		MOVWF 	OpA0		;Wenn nein: OpA0=W
		INCF  	Mx,1		;Mx erhoehen
		GOTO 	$-6
		MOVLW 	0x30		;Rueckgabe 0x30+Mx (im Zeichensatz ist "0" an Position 0x30 bis "9" an Position 0x39), OpA0 behaelt letzten Wert
		ADDWF 	Mx,0
		RETURN


;;****** C1vA1 *******************************************************
;Funktion: Extrahiert den 1er Character aus OpA0 und gibt diesen als Chr-Code fr Displays aus
;Funktion: W=CHRCODE(OpA[1]) : User muss darauf achten, dass OpA0 nicht grer als 9 ist
;Bentigte Subroutinen: keine
;Bentigte Variablen: OpA[1]
;Bentigte Register: keine
;Unverndert bleibt: OpA[1]
;Verndert werden  : keine
C1vA1		MOVLW  	0x30			
		ADDWF  	OpA0,0
		RETURN
	

;;******* DvEAB222 *******************************************************
;Funktion: Dividiert einen 16bit Wert durch einen 16bit Wert zu einem 16bit Ergebnis . Erg=OpA/OpB, OpA=Rest
;Funktion: Erg[2]=OpA[2]/OpB[2] , OpA[2]=Rest
;Bentigte Subroutinen: VvMB22, SvMMB222
;Bentigte Variablen: Erg[2], OpA[2], OpB[2], My, M[2]
;Bentigte Register: STATUS
;Unverndert bleibt: OpB[2]
;Verndert werden  : Erg[2], OpA[2], My, M[2], STATUS
DvEAB222	CLRF	M0			;Clear Zwischenergenis		
		CLRF	M1
		CLRF	Erg0		;Clear Endergebnis
		CLRF	Erg1
		MOVF	OpB0,0		;Prfe OpB0
		BTFSS	STATUS,2	;Pruefe, ob OpB0=0
		GOTO	$+8			;Nein: Beginne Division
		MOVF	OpB1,0		;Ja: Prfe OpB1
		BTFSS	STATUS,2	;Pruefe, ob OpB1=0
		GOTO	$+5			;Nein: Beginne Division
		MOVLW	0xFF		;Ja: Erg = FF:FF
		MOVWF	Erg0
		MOVWF	Erg1
		RETLW	.1
		MOVLW	.16			;Durch alle 16 Bits OpA gehen
		MOVWF	My
		BCF		STATUS,0	;Left-Shift Ergebnis	
		RLF		Erg0,1				
		RLF		Erg1,1	
		BCF		STATUS,0	;Left-Shift Zwischenwert	
		RLF		M0,1				
		RLF		M1,1	
		BCF		STATUS,0	;Left-Shift OpA nach Carry	
		RLF		OpA0,1				
		RLF		OpA1,1		
		BTFSC	STATUS,0	;Prfe eben geshiftetes Bit
		BSF		M0,0		;Ist ne 1: M0 Bit 0 auf 1
		CALL	VvMB22		;Vergleicht M mit OpB, STATUS,Carry (1 wenn M>=OpB, 0 wenn M<OpB)
		BTFSS	STATUS,0	;Pruefe Ergebnis
		GOTO	$+3			;M<OpB: nchstes OpA-Bit
		CALL	SvMMB222	;;M>=OpB: Subtraktion M=M-OpB
		BSF		Erg0,0		;Bit 0 von Erglow setzen
		DECFSZ	My,1		;Schleifenzaehler dekrement
		GOTO	$-17		;Noch keine 16 mal durchlaufen
		MOVF	M0,0		;16 mal durchlaufen: OpA=Rest
		MOVWF	OpA0
		MOVF	M1,0
		MOVWF	OpA1
		RETLW	.1
		

;;******* VvMB22 *******************************************************
;Funktion: Vergleicht einen 16bit Wert mit einem 16bit Wert.
;Funktion: CARRY=(M[2]>=OpB[2]) : CARRY = (1 wenn M>=OpB, 0 wenn M<OpB)
;Bentigte Subroutinen: keine
;Bentigte Variablen: M[2], OpB[2]
;Bentigte Register: STATUS
;Unverndert bleibt: M[2], OpB[2]
;Verndert werden  : STATUS
VvMB22	MOVF	OpB1,0		;Vergleiche OpB1 mit M1
		SUBWF	M1,0		;W=M1-OpB1
		BTFSS	STATUS,0	;Pruefe auf Ueberlauf
		GOTO	$+9			;Ja, Ueberauf. D.h. OpB1>M1. D.h. Ausgabe =0		
		BTFSS	STATUS,2	;Nein, OpB1<=M1: Pruefe, ob M1=OpB1
		GOTO	$+5			;M1 ist nicht gleich OpB1 -> M1>OpB1 -> Ausgabe =1
		MOVF	OpB0,0		;M1=OpB1, weitere Untersuchung auf Lowbytes
		SUBWF	M0,0		;W=M0-OpB0
		BTFSS	STATUS,0	;Pruefe auf Ueberlauf		
		GOTO	$+3			;Ja, Ueberauf. D.h. OpB0>M0. D.h. Ausgabe =0
		BSF		STATUS,0	;Nein, M0>=OpB0 D.h. Ergebnis=1
		RETURN
		BCF		STATUS,0
		RETURN		


;;******* SvMMB222 *******************************************************
;Funktion: Subtrahiert 16bit von 16bit zu 16bit Ergebnis.
;Funktion: M[2]=M[2]-OpB[2]
;Bentigte Subroutinen: keine
;Bentigte Variablen: M[2], OpB[2]
;Bentigte Register: STATUS
;Unverndert bleibt: OpB[2]
;Verndert werden  : M[2], STATUS
SvMMB222	MOVF	OpB0,0		;W=OpB0
		SUBWF	M0,1		;M0=M0-OpB0
		BTFSS	STATUS,0	;Pruefe, ob Ueberlauf
		DECF	M1,1		;Ueberlauf: M1=M1-1
		MOVF	OpB1,0		;Kein berlauf: W=OpB1
		SUBWF	M1,1		;M1=M1-OpB1
		RETURN		
		


;;******* SvEEA221 *******************************************************
;Funktion: Subtrahiert 8bit von 16bit zu 16bit Ergebnis.
;Funktion: Erg[2]=Erg[2]-OpA[1]
;Bentigte Subroutinen: keine
;Bentigte Variablen: Erg[2], OpA[1]
;Bentigte Register: STATUS
;Unverndert bleibt: OpA[1]
;Verndert werden  : Erg[2], STATUS
SvEEA221	MOVF	OpA0,0		;W=OpA0
		SUBWF	Erg0,1		;Erg0=Erg0-OpA0
		BTFSS	STATUS,0	;Pruefe, ob Ueberlauf
		DECF	Erg1,1		;Ueberlauf: Erg1=Erg1-1
		RETURN		
		
